# License: BSD 3 clause
'''
This package provides a number of utilities to make it simpler to run
common scikit-learn experiments with pre-generated features.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
'''

from __future__ import absolute_import, print_function, unicode_literals

from sklearn.metrics import f1_score, make_scorer, SCORERS

from .data import convert_examples, load_examples, write_feature_file
from .experiments import run_ablation, run_configuration
from .learner import Learner
from .metrics import (kappa, kendall_tau, spearman, pearson,
                      f1_score_least_frequent)
from .version import __version__, VERSION


__all__ = ['Learner', 'load_examples', 'kappa', 'kendall_tau', 'spearman',
           'pearson', 'f1_score_least_frequent', 'run_configuration',
           'run_ablation', 'write_feature_file', 'convert_examples']

# Add our scorers to the sklearn dictionary here so that they will always be
# available if you import anything from skll
_scorers = {'f1_score_micro': make_scorer(f1_score, average='micro', pos_label=None),
            'f1_score_macro': make_scorer(f1_score, average='macro', pos_label=None),
            'f1_score_weighted': make_scorer(f1_score, average='weighted', pos_label=None),
            'f1_score_least_frequent': make_scorer(f1_score_least_frequent),
            'pearson': make_scorer(pearson),
            'spearman': make_scorer(spearman),
            'kendall_tau': make_scorer(kendall_tau),
            'unweighted_kappa': make_scorer(kappa),
            'quadratic_weighted_kappa': make_scorer(kappa, weights='quadratic'),
            'linear_weighted_kappa': make_scorer(kappa, weights='linear'),
            'qwk_off_by_one': make_scorer(kappa, weights='quadratic',
                                          allow_off_by_one=True),
            'lwk_off_by_one': make_scorer(kappa, weights='linear',
                                          allow_off_by_one=True),
            'uwk_off_by_one': make_scorer(kappa, allow_off_by_one=True)}

SCORERS.update(_scorers)



#!/usr/bin/env python

# Copyright (C) 2012-2014 Educational Testing Service

# This file is part of SciKit-Learn Lab.

# SciKit-Learn Lab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# SciKit-Learn Lab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with SciKit-Learn Lab.  If not, see <http://www.gnu.org/licenses/>.

'''
script for computing additional evaluation metrics

:author: Michael Heilman (mheilman@ets.org)
'''
from __future__ import print_function, unicode_literals

import argparse
import csv
import logging

from skll.data import load_examples
from skll.metrics import use_score_func
from skll.version import __version__


def compute_eval_from_predictions(examples_file, predictions_file,
                                  metric_names):
    '''
    Compute evaluation metrics from prediction files after you have run an
    experiment.

    :param examples_file: a SKLL examples file (in .jsonlines or other format)
    :param predictions_file: a SKLL predictions output TSV file with id
                             and prediction column names
    :param metric_names: a list of SKLL metric names
                         (e.g., [pearson, unweighted_kappa])

    :returns: a dictionary from metrics names to values
    '''

    # read gold standard labels
    data = load_examples(examples_file)
    gold = dict(zip(data.ids, data.classes))

    # read predictions
    pred = {}
    with open(predictions_file) as pred_file:
        reader = csv.reader(pred_file, dialect=csv.excel_tab)
        next(reader)  # skip header
        for row in reader:
            pred[row[0]] = float(row[1])

    # make a sorted list of example ids in order to match up
    # labels and predictions
    if set(gold.keys()) != set(pred.keys()):
        raise ValueError('The example and prediction IDs do not match.')
    example_ids = sorted(gold.keys())

    res = {}
    for metric_name in metric_names:
        score = use_score_func(metric_name,
                               [gold[ex_id] for ex_id in example_ids],
                               [pred[ex_id] for ex_id in example_ids])
        res[metric_name] = score
    return res


def main(argv=None):
    '''
    Handles command line arguments and gets things started.

    :param argv: List of arguments, as if specified on the command-line.
                 If None, ``sys.argv[1:]`` is used instead.
    :type argv: list of str
    '''
    # Get command line arguments
    parser = argparse.ArgumentParser(description="Computes evaluation metrics \
                                                  from prediction files after \
                                                  you have run an experiment.",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('examples_file',
                        help='SKLL input file with labeled examples')
    parser.add_argument('predictions_file',
                        help='file with predictions from SKLL')
    parser.add_argument('metric_names',
                        help='metrics to compute',
                        nargs='+')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {0}'.format(__version__))
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format=('%(asctime)s - %(name)s - %(levelname)s - ' +
                                '%(message)s'))

    scores = compute_eval_from_predictions(args.examples_file,
                                           args.predictions_file,
                                           args.metric_names)

    for metric_name in args.metric_names:
        print("{}\t{}\t{}".format(scores[metric_name],
                                  metric_name, args.predictions_file))


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright (C) 2012-2013 Educational Testing Service

# This file is part of SciKit-Learn Lab.

# SciKit-Learn Lab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# SciKit-Learn Lab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with SciKit-Learn Lab.  If not, see <http://www.gnu.org/licenses/>.

'''
Runs a bunch of scikit-learn jobs in parallel on the cluster given a
config file.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
'''


from __future__ import print_function, unicode_literals

import argparse
import logging
import sys
from functools import partial

from skll.experiments import run_configuration
from skll.version import __version__


def main(argv=None):
    '''
    Handles command line arguments and gets things started.

    :param argv: List of arguments, as if specified on the command-line.
                 If None, ``sys.argv[1:]`` is used instead.
    :type argv: list of str
    '''
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Runs the scikit-learn experiments in a given config file.\
                     If Grid Map is installed, jobs will automatically be \
                     created and run on a DRMAA-compatible cluster.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler='resolve')
    parser.add_argument('config_file',
                        help='Configuration file describing the sklearn task\
                              to run.',
                        nargs='+')
    parser.add_argument('-a', '--ablation',
                        help='Runs an ablation study where repeated \
                              experiments are conducted where the specified \
                              number of features in each featureset in the \
                              configuration file are held out.',
                        type=int, default=0,
                        metavar='NUM_FEATURES')
    parser.add_argument('-A', '--ablation_all',
                        help='Runs an ablation study where repeated \
                              experiments are conducted with all combinations \
                              of features in each featureset in the \
                              configuration file. Overrides --ablation \
                              setting.',
                        action='store_true')
    parser.add_argument('-k', '--keep_models',
                        help='If trained models already exists, re-use them\
                              instead of overwriting them.',
                        action='store_true')
    parser.add_argument('-l', '--local',
                        help='Do not use the Grid Engine for running jobs and\
                              just run everything sequentially on the local \
                              machine. This is for debugging.',
                        action='store_true')
    parser.add_argument('-m', '--machines',
                        help="comma-separated list of machines to add to\
                              gridmap's whitelist (if not specified, all\
                              available machines are used). Note that full \
                              names must be specified, e.g., \
                              \"nlp.research.ets.org\"",
                        default=None)
    parser.add_argument('-q', '--queue',
                        help="Use this queue for gridmap.",
                        default='all.q')
    parser.add_argument('-r', '--resume',
                        help='If result files already exist for an experiment, \
                              do not overwrite them. This is very useful when \
                              doing a large ablation experiment and part of it \
                              crashes.',
                        action='store_true')
    parser.add_argument('-v', '--verbose',
                        help='Print more status information. For every ' +
                             'additional time this flag is specified, ' +
                             'output gets more verbose.',
                        default=0, action='count')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {0}'.format(__version__))
    args = parser.parse_args(argv)

    # Logging levels are really integer multiples of 10, so convert verbose flag
    log_level = max(logging.WARNING - (args.verbose * 10), logging.DEBUG)
    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format=('%(asctime)s - %(name)s - %(levelname)s - ' +
                                '%(message)s'), level=log_level)

    machines = None
    if args.machines:
        machines = args.machines.split(',')

    # Create partial function to map onto list of config files
    ablation = args.ablation
    if args.ablation_all:
        ablation = None

    # Run each config file sequentially
    for config_file in args.config_file:
        run_configuration(config_file, local=args.local, overwrite=not
                          args.keep_models, queue=args.queue, hosts=machines,
                          ablation=ablation, resume=args.resume)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright (C) 2012-2013 Educational Testing Service

# This file is part of SciKit-Learn Lab.

# SciKit-Learn Lab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# SciKit-Learn Lab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with SciKit-Learn Lab.  If not, see <http://www.gnu.org/licenses/>.

"""
Join MegaM files

:author: Dan Blanchard (dblanchard@ets.org)
:date: July 2012
"""

from __future__ import print_function, unicode_literals

import argparse
import logging
import os
import re
import sys
from collections import OrderedDict

from six import iteritems

from skll.data import _MegaMDictIter
from skll.version import __version__


# Globals
warned_about = set()


def parse_num_list(num_string):
    '''
    Convert a string representing a range of numbers to a list of integers.
    '''
    range_list = []
    if (num_string != '') and (not re.match(r'^(\d+(-\d+)?,)*\d+(-\d+)?$',
                                            num_string)):
        raise argparse.ArgumentTypeError("'" + num_string + "' is not a range" +
                                         " of numbers. Expected forms are " +
                                         "'8-15', '4,8,15,16,23,42', or " +
                                         "'8-15,42'.")
    for rng in num_string.split(','):
        if rng.count('-'):
            split_range = [int(x) for x in rng.split('-')]
            split_range[1] += 1
            range_list.extend(range(*split_range))
        else:
            range_list.append(int(rng))
    return range_list


def get_unique_name(feature_name, prev_feature_set, filename):
    '''
    Get a name that doesn't overlap with the previous features.

    :param feature_name: The feature that we want a unique name for.
    :type feature_name: string
    :param prev_feature_set: The feature names encountered so far.
    :type prev_feature_set: set of strings
    :param filename: The current MegaM file we're processing.
    :type filename: string

    :returns: Either feature_name or feature_name with a unique suffix based
              on the current MegaM file if there was an overlap.
    '''
    global warned_about

    new_feature_name = feature_name
    suffix = os.path.splitext(os.path.basename(filename))[0].replace(' ', '_')
    # Add suffix multiple times if necessary
    while new_feature_name in prev_feature_set:
        new_feature_name += "_" + suffix

    logger = logging.getLogger(__name__)
    if new_feature_name != feature_name and feature_name not in warned_about:
        logger.warning(("Feature named {} already found in previous files. " +
                        "Renaming to {} to prevent " +
                        "duplicates.").format(feature_name,
                                              new_feature_name).encode('utf-8'))
        warned_about.add(feature_name)
    return new_feature_name


def main(argv=None):
    '''
    Handles command line arguments and gets things started.

    :param argv: List of arguments, as if specified on the command-line.
                 If None, ``sys.argv[1:]`` is used instead.
    :type argv: list of str
    '''
    # Get command line arguments
    parser = argparse.ArgumentParser(description="Combine MegaM files that \
                                                  contain features for the same\
                                                  files.",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('megam_file',
                        help='MegaM input file(s). Each feature line must be \
                              preceded by a comment with the filename/ID that \
                              the features should be joined on.',
                        nargs='+')
    parser.add_argument('-b', '--binary',
                        help='Converts all of the features in the specified \
                              range of files to presence/absence binary \
                              features. Files are numbered starting from 1, and\
                              if 0 is specified with this flag, all files are\
                              converted.',
                        type=parse_num_list)
    parser.add_argument('--doubleup',
                        help='Keep both the binary and numeric versions of any\
                              feature you convert to binary.',
                        action='store_true')
    parser.add_argument('-c', '--common',
                        help='Only output features for filenames that are \
                              common to all MegaM files.',
                        action='store_true')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {0}'.format(__version__))
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format=('%(asctime)s - %(name)s - %(levelname)s - ' +
                                '%(message)s'))
    logger = logging.getLogger(__name__)

    # Map from filenames to feature strings
    feature_dict = OrderedDict()
    class_dict = {}
    filename_set = set()

    # Set that will contain all of the features seen in previous files
    # (for duplicate detection)
    prev_feature_set = set()

    # Iterate through MegaM files
    for file_num, infile in enumerate(args.megam_file, start=1):
        # Initialize duplicate feature book-keeping variables
        curr_feature_set = set()

        # Initialize set for storing filenames mentioned in current MegaM file
        curr_filename_set = set()

        # Handle current MegaM file
        for curr_filename, class_name, feature_dict in _MegaMDictIter(infile):
            if curr_filename in class_dict:
                if class_dict[curr_filename] != class_name:
                    raise ValueError(("Inconsisten class label for instance " +
                                      "{} in {}.").format(curr_filename,
                                                          infile.name))
            else:
                class_dict[curr_filename] = class_name
            # If there are non-zero features, process them
            if feature_dict:
                for feat_name, feat_val in iteritems(feature_dict):
                    # Handle duplicate features
                    feat_name = get_unique_name(feat_name, prev_feature_set,
                                                infile.name)
                    # Ignore zero-valued features
                    try:
                        if feat_val != 'N/A' and float(feat_val) != 0:
                            # Convert feature to binary if necessary
                            if (args.binary and ((args.binary == [0]) or (file_num in args.binary))):
                                if args.doubleup:
                                    new_feat_pair = '{} {} '.format(feat_name, feat_val)
                                    feature_dict[curr_filename] = new_feat_pair if curr_filename not in feature_dict else feature_dict[curr_filename] + new_feat_pair
                                    curr_feature_set.add(feat_name)
                                    feat_name = get_unique_name(feat_name + "_binary", prev_feature_set, infile.name)
                                feat_val = 1

                            # Add feature pair to current string of features
                            new_feat_pair = '{} {} '.format(feat_name,
                                                            feat_val)
                            feature_dict[curr_filename] = new_feat_pair if curr_filename not in feature_dict else feature_dict[curr_filename] + new_feat_pair
                            curr_feature_set.add(feat_name)
                    except ValueError:
                        raise ValueError(("Invalid feature value in feature " +
                                          "pair '{} {}' for file {}").format(feat_name,
                                                                             feat_val,
                                                                             curr_filename).encode('utf-8'))

            # Otherwise warn about lack of features (although that really
            # just means all of them have zero values)
            else:
                if curr_filename not in feature_dict:
                    feature_dict[curr_filename] = ""
                logger.warning(("No features found for {} in {}. All are " +
                                "assumed to be zero.").format(curr_filename,
                                                              infile.name).encode('utf-8'))

        # Add current file's features to set of seen features
        prev_feature_set.update(curr_feature_set)

        # Either intersect or union current file's filenames with existing ones
        if args.common and filename_set:
            filename_set.intersection_update(curr_filename_set)
        else:
            filename_set.update(curr_filename_set)

    # Print new MegaM file
    for curr_filename in feature_dict.viewkeys():
        # Skip files that aren't common when args.common is true
        if curr_filename not in filename_set:
            continue
        print("# {}".format(curr_filename).encode('utf-8'))
        print("{}\t{}".format(class_dict[curr_filename],
                              feature_dict[curr_filename].strip()).encode('utf-8'))


if __name__ == '__main__':
    main()


#!/usr/bin/env python
# License: BSD 3 clause
'''
Little helper script to create a summary file out of a list of JSON results
files.

:author: Dan Blanchard (dblanchard@ets.org)
:date: September 2013
'''

from __future__ import print_function, unicode_literals

import argparse
import logging
import sys
from io import open

from skll.experiments import _write_summary_file
from skll.version import __version__


def main(argv=None):
    '''
    Handles command line arguments and gets things started.

    :param argv: List of arguments, as if specified on the command-line.
                 If None, ``sys.argv[1:]`` is used instead.
    :type argv: list of str
    '''
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Creates an experiment summary TSV file from a list of JSON\
                     files generated by run_experiment.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler='resolve')
    parser.add_argument('summary_file',
                        help='TSV file to store summary of results.')
    parser.add_argument('json_file',
                        help='JSON results file generated by run_experiment.',
                        nargs='+')
    parser.add_argument('-a', '--ablation', action='store_true', default=False,
                        help='The results files are from an ablation run.')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {0}'.format(__version__))
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format=('%(asctime)s - %(name)s - %(levelname)s - ' +
                                '%(message)s'))

    file_mode = 'w' if sys.version_info >= (3, 0) else 'wb'
    with open(args.summary_file, file_mode) as output_file:
        _write_summary_file(args.json_file, output_file,
                            ablation=int(args.ablation))


if __name__ == '__main__':
    main()



# License: BSD 3 clause
'''
Handles loading data from various types of data files.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
'''

from __future__ import print_function, unicode_literals

import csv
import json
import logging
import os
import re
import sys
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from csv import DictReader, DictWriter
from decimal import Decimal
from itertools import chain, islice
from io import open, BytesIO, StringIO
from multiprocessing import Queue
from operator import itemgetter

import numpy as np
from bs4 import UnicodeDammit
from collections import namedtuple
from six import iteritems, string_types, text_type
from six.moves import map, zip
from sklearn.feature_extraction import DictVectorizer, FeatureHasher

# Import QueueHandler and QueueListener for multiprocess-safe logging
if sys.version_info < (3, 0):
    from logutils.queue import QueueHandler, QueueListener
else:
    from logging.handlers import QueueHandler, QueueListener

MAX_CONCURRENT_PROCESSES = int(os.getenv('SKLL_MAX_CONCURRENT_PROCESSES', '5'))


ExamplesTuple = namedtuple('ExamplesTuple', ['ids', 'classes', 'features',
                                             'feat_vectorizer'])

# Register dialect for handling ARFF files
if sys.version_info >= (3, 0):
    csv.register_dialect('arff', delimiter=',', quotechar="'",
                         escapechar='\\', doublequote=False,
                         lineterminator='\n', skipinitialspace=True)
else:
    csv.register_dialect('arff', delimiter=b',', quotechar=b"'",
                         escapechar=b'\\', doublequote=False,
                         lineterminator=b'\n', skipinitialspace=True)


class _DictIter(object):
    """
    A little helper class to make picklable iterators out of example
    dictionary generators

    :param path_or_list: Path or a list of example dictionaries.
    :type path_or_list: str or list of dict
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    :param class_map: Mapping from original class labels to new ones. This is
                      mainly used for collapsing multiple classes into a single
                      class. Anything not in the mapping will be kept the same.
    :type class_map: dict from str to str
    """
    def __init__(self, path_or_list, quiet=True, ids_to_floats=False,
                 label_col='y', class_map=None):
        super(_DictIter, self).__init__()
        self.path_or_list = path_or_list
        self.quiet = quiet
        self.ids_to_floats = ids_to_floats
        self.label_col = label_col
        self.class_map = class_map

    def __iter__(self):
        # Setup logger
        logger = logging.getLogger(__name__)

        logger.debug('DictIter type: %s', type(self))
        logger.debug('path_or_list: %s', self.path_or_list)

        # Check if we're given a path to a file, and if so, open it
        if isinstance(self.path_or_list, string_types):
            if sys.version_info >= (3, 0):
                file_mode = 'r'
            else:
                file_mode = 'rb'
            f = open(self.path_or_list, file_mode)
            if not self.quiet:
                print("Loading {}...".format(self.path_or_list), end="",
                      file=sys.stderr)
                sys.stderr.flush()
            return self._sub_iter(f)
        else:
            if not self.quiet:
                print("Loading...".format(self.path_or_list), end="",
                      file=sys.stderr)
                sys.stderr.flush()
            return self._sub_iter(self.path_or_list)

    def _sub_iter(self, file_or_list):
        '''
        Iterates through a given file or list.
        '''
        raise NotImplementedError


class _DummyDictIter(_DictIter):
    '''
    This class is to facilitate programmatic use of ``Learner.predict()``
    and other functions that take ``ExamplesTuple`` objects as input.
    It iterates over examples in the same way as other ``_DictIter``s, but uses
    a list of example dictionaries instead of a path to a file.

    :param path_or_list: List of example dictionaries.
    :type path_or_list: list of dict
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    '''
    def __iter__(self):
        if not self.quiet:
            print("Converting examples...", end="", file=sys.stderr)
            sys.stderr.flush()
        for example_num, example in enumerate(self.path_or_list):
            curr_id = str(example.get("id",
                                      "EXAMPLE_{}".format(example_num)))
            if self.ids_to_floats:
                try:
                    curr_id = float(curr_id)
                except ValueError:
                    raise ValueError(('You set ids_to_floats to true,' +
                                      ' but ID {} could not be ' +
                                      'converted to in ' +
                                      '{}').format(curr_id,
                                                   example))
            class_name = (_safe_float(example['y'], replace_dict=self.class_map)
                          if 'y' in example else None)
            example = example['x']
            yield curr_id, class_name, example

            if not self.quiet and example_num % 100 == 0:
                print(".", end="", file=sys.stderr)
        if not self.quiet:
            print("done", file=sys.stderr)


class _JSONDictIter(_DictIter):
    '''
    Iterator to convert current line in .jsonlines/.ndj file to a dictionary
    with the following fields: "SKLL_ID" (originally "id"), "SKLL_CLASS_LABEL"
    (originally "y"), and all of the feature-values in the sub-dictionary "x".
    Basically, we're flattening the structure, but renaming "y" and "id" to
    prevent possible conflicts with feature names in "x".

    :param path_or_list: Path to .jsonlines/.ndj file
    :type path_or_list: str
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    '''

    def _sub_iter(self, file_or_list):
        for example_num, line in enumerate(file_or_list):
            # Remove extraneous whitespace
            line = line.strip()

            # If this is a comment line or a blank line, move on
            if line.startswith('//') or not line:
                continue

            # Process good lines
            example = json.loads(line)
            # Convert all IDs to strings initially,
            # for consistency with csv and megam formats.
            curr_id = str(example.get("id",
                                      "EXAMPLE_{}".format(example_num)))
            class_name = (_safe_float(example['y'], replace_dict=self.class_map)
                          if 'y' in example else None)
            example = example["x"]

            if self.ids_to_floats:
                try:
                    curr_id = float(curr_id)
                except ValueError:
                    raise ValueError(('You set ids_to_floats to true, but' +
                                      ' ID {} could not be converted to ' +
                                      'float').format(curr_id))

            yield curr_id, class_name, example

            if not self.quiet and example_num % 100 == 0:
                print(".", end="", file=sys.stderr)
        if not self.quiet:
            print("done", file=sys.stderr)


class _MegaMDictIter(_DictIter):
    '''
    Iterator that yields tuples of IDs, classes, and dictionaries mapping from
    features to values for each pair of lines in the MegaM -fvals file specified
    by path.

    :param path_or_list: Path to .megam file (-fvals format)
    :type path_or_list: str
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    '''

    def _sub_iter(self, file_or_list):
        example_num = 0
        curr_id = 'EXAMPLE_0'
        for line in file_or_list:
            # Process encoding
            if not isinstance(line, text_type):
                line = UnicodeDammit(line, ['utf-8',
                                            'windows-1252']).unicode_markup
            line = _sanitize_line(line.strip())
            # Handle instance lines
            if line.startswith('#'):
                curr_id = line[1:].strip()
            elif line and line not in ['TRAIN', 'TEST', 'DEV']:
                split_line = line.split()
                num_cols = len(split_line)
                del line
                # Line is just a class label
                if num_cols == 1:
                    class_name = _safe_float(split_line[0],
                                             replace_dict=self.class_map)
                    field_pairs = []
                # Line has a class label and feature-value pairs
                elif num_cols % 2 == 1:
                    class_name = _safe_float(split_line[0],
                                             replace_dict=self.class_map)
                    field_pairs = split_line[1:]
                # Line just has feature-value pairs
                elif num_cols % 2 == 0:
                    class_name = None
                    field_pairs = split_line

                curr_info_dict = {}
                if len(field_pairs) > 0:
                    # Get current instances feature-value pairs
                    field_names = islice(field_pairs, 0, None, 2)
                    # Convert values to floats, because otherwise
                    # features'll be categorical
                    field_values = (_safe_float(val) for val in
                                    islice(field_pairs, 1, None, 2))

                    # Add the feature-value pairs to dictionary
                    curr_info_dict.update(zip(field_names, field_values))

                    if len(curr_info_dict) != len(field_pairs) / 2:
                        raise ValueError(('There are duplicate feature ' +
                                          'names in {} for example ' +
                                          '{}.').format(self.path_or_list,
                                                        curr_id))

                if self.ids_to_floats:
                    try:
                        curr_id = float(curr_id)
                    except ValueError:
                        raise ValueError(('You set ids_to_floats to true,' +
                                          ' but ID {} could not be ' +
                                          'converted to in ' +
                                          '{}').format(curr_id,
                                                       self.path_or_list))

                yield curr_id, class_name, curr_info_dict

                # Set default example ID for next instance, in case we see a
                # line without an ID.
                example_num += 1
                curr_id = 'EXAMPLE_{}'.format(example_num)

                if not self.quiet and example_num % 100 == 0:
                    print(".", end="", file=sys.stderr)
        if not self.quiet:
            print("done", file=sys.stderr)


class _DelimitedDictIter(_DictIter):
    '''
    Iterator that yields tuples of IDs, classes, and dictionaries mapping from
    features to values for each line in a specified delimited (CSV/TSV) file.

    :param path_or_list: Path to .tsv file
    :type path_or_list: str
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param label_col: Name of the column which contains the class labels.
                      If no column with that name exists, or `None` is
                      specified, the data is considered to be unlabelled.
    :type label_col: str
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    :param class_map: Mapping from original class labels to new ones. This is
                      mainly used for collapsing multiple classes into a single
                      class. Anything not in the mapping will be kept the same.
    :type class_map: dict from str to str
    :param dialect: The dialect of to pass on to the underlying CSV reader.
    :type dialect: str
    '''
    def __init__(self, path_or_list, quiet=True, ids_to_floats=False,
                 label_col='y', class_map=None, dialect=None):
        super(_DelimitedDictIter, self).__init__(path_or_list, quiet=quiet,
                                                 ids_to_floats=ids_to_floats,
                                                 label_col=label_col,
                                                 class_map=class_map)
        self.dialect = dialect

    def _sub_iter(self, file_or_list):
        reader = DictReader(file_or_list, dialect=self.dialect)
        for example_num, row in enumerate(reader):
            if self.label_col is not None and self.label_col in row:
                class_name = _safe_float(row[self.label_col],
                                         replace_dict=self.class_map)
                del row[self.label_col]
            else:
                class_name = None

            if "id" not in row:
                curr_id = "EXAMPLE_{}".format(example_num)
            else:
                curr_id = row["id"]
                del row["id"]

            # Convert features to floats and if a feature is 0
            # then store the name of the feature so we can
            # delete it later since we don't need to explicitly
            # store zeros in the feature hash
            columns_to_delete = []
            if sys.version_info < (3, 0):
                columns_to_convert_to_unicode = []
            for fname, fval in iteritems(row):
                fval_float = _safe_float(fval)
                # we don't need to explicitly store zeros
                if fval_float:
                    row[fname] = fval_float
                    if sys.version_info < (3, 0):
                        columns_to_convert_to_unicode.append(fname)
                else:
                    columns_to_delete.append(fname)

            # remove the columns with zero values
            for cname in columns_to_delete:
                del row[cname]

            # convert the names of all the other columns to
            # unicode for python 2
            if sys.version_info < (3, 0):
                for cname in columns_to_convert_to_unicode:
                    fval = row[cname]
                    del row[cname]
                    row[cname.decode('utf-8')] = fval

            if self.ids_to_floats:
                try:
                    curr_id = float(curr_id)
                except ValueError:
                    raise ValueError(('You set ids_to_floats to true, but' +
                                      ' ID {} could not be converted to ' +
                                      'float').format(curr_id))
            elif sys.version_info < (3, 0):
                curr_id = curr_id.decode('utf-8')

            yield curr_id, class_name, row

            if not self.quiet and example_num % 100 == 0:
                print(".", end="", file=sys.stderr)
        if not self.quiet:
            print("done", file=sys.stderr)


class _CSVDictIter(_DelimitedDictIter):
    '''
    Iterator that yields tuples of IDs, classes, and dictionaries mapping from
    features to values for each line in a specified CSV file.

    :param path_or_list: Path to .tsv file
    :type path_or_list: str
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param label_col: Name of the column which contains the class labels.
                      If no column with that name exists, or `None` is
                      specified, the data is considered to be unlabelled.
    :type label_col: str
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    '''
    def __init__(self, path_or_list, quiet=True, ids_to_floats=False,
                 label_col='y', class_map=None):
        super(_CSVDictIter, self).__init__(path_or_list, quiet=quiet,
                                           ids_to_floats=ids_to_floats,
                                           label_col=label_col,
                                           dialect='excel',
                                           class_map=class_map)


class _ARFFDictIter(_DelimitedDictIter):
    '''
    Iterator that yields tuples of IDs, classes, and dictionaries mapping from
    features to values for each data line in a specified ARFF file.

    :param path_or_list: Path to .tsv file
    :type path_or_list: str
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param label_col: Name of the column which contains the class labels.
                      If no column with that name exists, or `None` is
                      specified, the data is considered to be unlabelled.
    :type label_col: str
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    '''
    def __init__(self, path_or_list, quiet=True, ids_to_floats=False,
                 label_col='y', class_map=None):
        super(_ARFFDictIter, self).__init__(path_or_list, quiet=quiet,
                                            ids_to_floats=ids_to_floats,
                                            label_col=label_col,
                                            dialect='arff',
                                            class_map=class_map)

    @staticmethod
    def split_with_quotes(s, delimiter=' ', quote_char="'", escape_char='\\'):
        '''
        A replacement for string.split that won't split delimiters enclosed in
        quotes.
        '''
        if sys.version_info < (3, 0):
            delimiter = delimiter.encode()
            quote_char = quote_char.encode()
            escape_char = escape_char.encode()
        return next(csv.reader([s], delimiter=delimiter, quotechar=quote_char,
                               escapechar=escape_char))

    def _sub_iter(self, file_or_list):
        field_names = []
        # Process ARFF header
        for line in file_or_list:
            # Process encoding
            if not isinstance(line, text_type):
                decoded_line = UnicodeDammit(line, ['utf-8',
                                             'windows-1252']).unicode_markup
            else:
                decoded_line = line
            line = decoded_line.strip()
            # Skip empty lines
            if line:
                # Split the line using CSV reader because it can handle
                # quoted delimiters.
                split_header = self.split_with_quotes(line)
                row_type = split_header[0].lower()
                if row_type == '@attribute':
                    # Add field name to list
                    field_names.append(split_header[1])
                # Stop at data
                elif row_type == '@data':
                    break
                # Skip other types of rows (relations)

        # Create header for CSV
        if sys.version_info < (3, 0):
            io_type = BytesIO
        else:
            io_type = StringIO
        with io_type() as field_buffer:
            csv.writer(field_buffer, dialect='arff').writerow(field_names)
            field_str = field_buffer.getvalue()

        # Set label_col to be the name of the last field, since that's standard
        # for ARFF files
        if self.label_col != field_names[-1]:
            self.label_col = None

        # Process data as CSV file
        return super(_ARFFDictIter, self)._sub_iter(chain([field_str],
                                                          file_or_list))


class _TSVDictIter(_DelimitedDictIter):
    '''
    Iterator that yields tuples of IDs, classes, and dictionaries mapping from
    features to values for each line in a specified TSV file.

    :param path_or_list: Path to .tsv file
    :type path_or_list: str
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param label_col: Name of the column which contains the class labels.
                      If no column with that name exists, or `None` is
                      specified, the data is considered to be unlabelled.
    :type label_col: str
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    '''
    def __init__(self, path_or_list, quiet=True, ids_to_floats=False,
                 label_col='y', class_map=None):
        super(_TSVDictIter, self).__init__(path_or_list, quiet=quiet,
                                           ids_to_floats=ids_to_floats,
                                           label_col=label_col,
                                           dialect='excel-tab',
                                           class_map=class_map)


def _ids_for_iter_type(example_iter_type, path, ids_to_floats):
    '''
    Little helper function to return an array of IDs for a given example
    generator (and whether or not the examples have labels).
    '''
    try:
        example_iter = example_iter_type(path, ids_to_floats=ids_to_floats)
        res_array = np.array([curr_id for curr_id, _, _ in example_iter])
    except Exception as e:
        # Setup logger
        logger = logging.getLogger(__name__)
        logger.exception('Failed to load IDs for %s.', path)
        raise e
    return res_array


def _classes_for_iter_type(example_iter_type, path, label_col, class_map):
    '''
    Little helper function to return an array of classes for a given example
    generator (and whether or not the examples have labels).
    '''
    try:
        example_iter = example_iter_type(path, label_col=label_col,
                                         class_map=class_map)
        res_array = np.array([class_name for _, class_name, _ in example_iter])
    except Exception as e:
        # Setup logger
        logger = logging.getLogger(__name__)
        logger.exception('Failed to load classes for %s.', path)
        raise e
    return res_array


def _features_for_iter_type(example_iter_type, path, quiet, sparse, label_col,
                            feature_hasher, num_features):
    '''
    Little helper function to return a sparse matrix of features and feature
    vectorizer for a given example generator (and whether or not the examples
    have labels).
    '''
    try:
        example_iter = example_iter_type(path, quiet=quiet, label_col=label_col)
        if feature_hasher:
            feat_vectorizer = FeatureHasher(n_features=num_features)
        else:
            feat_vectorizer = DictVectorizer(sparse=sparse)
        feat_dict_generator = map(itemgetter(2), example_iter)
    except Exception as e:
        # Setup logger
        logger = logging.getLogger(__name__)
        logger.exception('Failed to load features for %s.', path)
        raise e
    try:
        if feature_hasher:
            features = feat_vectorizer.transform(feat_dict_generator)
        else:
            features = feat_vectorizer.fit_transform(feat_dict_generator)
    except ValueError:
        raise ValueError('The last feature file did not include any features.')
    return features, feat_vectorizer


def load_examples(path, quiet=False, sparse=True, label_col='y',
                  ids_to_floats=False, class_map=None, feature_hasher=False,
                  num_features=None):
    '''
    Loads examples in the ``.arff``, ``.csv``, ``.jsonlines``, ``.megam``,
    ``.ndj``, or ``.tsv`` formats.

    If you would like to include example/instance IDs in your files, they must
    be specified in the following ways:

    * MegaM: As a comment line directly preceding the line with feature values.
    * CSV/TSV/ARFF: An "id" column.
    * JSONLINES: An "id" key in each JSON dictionary.

    Also, for ARFF, CSV, and TSV files, there must be a column with the name
    specified by `label_col` if the data is labelled. For ARFF files, this
    column must also be the final one (as it is in Weka).

    :param path: The path to the file to load the examples from, or a list of
                 example dictionaries (like you would pass to
                 `convert_examples`).
    :type path: str or dict
    :param quiet: Do not print "Loading..." status message to stderr.
    :type quiet: bool
    :param sparse: Whether or not to store the features in a numpy CSR matrix.
    :type sparse: bool
    :param label_col: Name of the column which contains the class labels for
                      ARFF/CSV/TSV files. If no column with that name exists, or
                      `None` is specified, the data is considered to be
                      unlabelled.
    :type label_col: str
    :param ids_to_floats: Convert IDs to float to save memory. Will raise error
                          if we encounter an a non-numeric ID.
    :type ids_to_floats: bool
    :param class_map: Mapping from original class labels to new ones. This is
                      mainly used for collapsing multiple classes into a single
                      class. Anything not in the mapping will be kept the same.
    :type class_map: dict from str to str

    :return: 4-tuple of an array example ids, an array of class labels, a
             scipy CSR matrix of features, and a DictVectorizer containing
             the mapping between feature names and the column indices in
             the feature matrix.
    '''
    # Setup logger
    logger = logging.getLogger(__name__)

    logger.debug('Path: %s', path)

    # Build an appropriate generator for examples so we process the input file
    # through the feature vectorizer without using tons of memory
    if not isinstance(path, string_types):
        example_iter_type = _DummyDictIter
    # Lowercase path for file extension checking, if it's a string
    else:
        lc_path = path.lower()
        if lc_path.endswith(".tsv"):
            example_iter_type = _TSVDictIter
        elif lc_path.endswith(".csv"):
            example_iter_type = _CSVDictIter
        elif lc_path.endswith(".arff"):
            example_iter_type = _ARFFDictIter
        elif lc_path.endswith(".jsonlines") or lc_path.endswith('.ndj'):
            example_iter_type = _JSONDictIter
        elif lc_path.endswith(".megam"):
            example_iter_type = _MegaMDictIter
        else:
            raise ValueError(('Example files must be in either .arff, .csv, ' +
                              '.jsonlines, .megam, .ndj, or .tsv format. You ' +
                              'specified: {}').format(path))

    logger.debug('Example iterator type: %s', example_iter_type)

    # Generators can't be pickled, so unfortunately we have to turn them into
    # lists. Would love a workaround for this.
    if example_iter_type == _DummyDictIter:
        path = list(path)

    if MAX_CONCURRENT_PROCESSES == 1:
        executor_type = ThreadPoolExecutor
    else:
        executor_type = ProcessPoolExecutor

    # Create thread/process-safe logger stuff
    queue = Queue(-1)
    q_handler = QueueHandler(queue)
    logger = logging.getLogger(__name__)
    logger.addHandler(q_handler)
    q_listener = QueueListener(queue)
    q_listener.start()

    # Create generators that we can use to create numpy arrays without wasting
    # memory (even though this requires reading the file multiple times).
    # Do this using a process pool so that we can clear out the temporary
    # variables more easily and do these in parallel.
    with executor_type(max_workers=3) as executor:
        ids_future = executor.submit(_ids_for_iter_type, example_iter_type,
                                     path, ids_to_floats)
        classes_future = executor.submit(_classes_for_iter_type,
                                         example_iter_type, path, label_col,
                                         class_map)
        features_future = executor.submit(_features_for_iter_type,
                                          example_iter_type, path, quiet,
                                          sparse, label_col, feature_hasher,
                                          num_features)
        # Wait for processes/threads to complete and store results
        ids = ids_future.result()
        classes = classes_future.result()
        features, feat_vectorizer = features_future.result()

    # Tear-down thread/process-safe logging and switch back to regular
    q_listener.stop()
    logger.removeHandler(q_handler)

    # Make sure we have the same number of ids, classes, and features
    assert ids.shape[0] == classes.shape[0] == features.shape[0]

    return ExamplesTuple(ids, classes, features, feat_vectorizer)


def convert_examples(example_dicts, sparse=True, ids_to_floats=False):
    '''
    This function is to facilitate programmatic use of Learner.predict()
    and other functions that take ExamplesTuple objects as input.
    It converts a .jsonlines/.ndj-style list of dictionaries into
    an ExamplesTuple.

    :param example_dicts: An list of dictionaries following the .jsonlines/.ndj
                          format (i.e., features 'x', label 'y', and 'id').
    :type example_dicts: iterable of dicts

    :return an ExamplesTuple representing the examples in example_dicts.
    '''
    return load_examples(example_dicts, sparse=sparse, quiet=True,
                         ids_to_floats=ids_to_floats)


def _sanitize_line(line):
    '''
    :param line: The line to clean up.
    :type line: str

    :returns: Copy of line with all non-ASCII characters replaced with
    <U1234> sequences where 1234 is the value of ord() for the character.
    '''
    char_list = []
    for char in line:
        char_num = ord(char)
        char_list.append('<U{}>'.format(char_num) if char_num > 127 else char)
    return ''.join(char_list)


def _safe_float(text, replace_dict=None):
    '''
    Attempts to convert a string to a float, but if that's not possible, just
    returns the original value.

    :param text: The text to convert.
    :type text: str
    :param replace_dict: Mapping from text to replacement text values. This is
                         mainly used for collapsing multiple classes into a
                         single class. Replacing happens before conversion to
                         floats. Anything not in the mapping will be kept the
                         same.
    :type replace_dict: dict from str to str
    '''
    if replace_dict is not None and text in replace_dict:
        text = replace_dict[text]
    try:
        return float(text)
    except ValueError:
        return text.decode('utf-8') if sys.version_info < (3, 0) else text
    except TypeError:
        return 0.0


def _examples_to_dictwriter_inputs(ids, classes, features, label_col):
    '''
    Takes lists of IDs, classes, and features and converts them to a set of
    fields and a list of instance dictionaries to use as input for DictWriter.
    '''
    instances = []
    fields = set()
    # Iterate through examples
    for ex_id, class_name, feature_dict in zip(ids, classes, features):
        # Don't try to add class column if this is label-less data
        if label_col not in feature_dict:
            if class_name is not None:
                feature_dict[label_col] = class_name
        else:
            raise ValueError(('Class column name "{0}" already used ' +
                              'as feature name!').format(label_col))

        if 'id' not in feature_dict:
            if ex_id is not None:
                feature_dict['id'] = ex_id
        else:
            raise ValueError('ID column name "id" already used as ' +
                             'feature name!')
        fields.update(feature_dict.keys())
        instances.append(feature_dict)
    return fields, instances


def _write_arff_file(path, ids, classes, features, label_col,
                     relation='skll_relation', arff_regression=False):
    '''
    Writes a feature file in .csv or .tsv format with the given a list of IDs,
    classes, and features.

    :param path: A path to the feature file we would like to create.
    :type path: str
    :param ids: The IDs for each instance in the feature list/array. If None,
                IDs will be automatically generated with the prefix specified by
                `id_prefix` followed by the row number. If `id_prefix` is also
                None, no IDs will be written to the file.
    :type ids: list of str
    :param classes: The class labels for each instance in the feature
                    list/array. If None, no class labels will be added to output
                    file.
    :type classes: list of str
    :param features: The features for each instance.
    :type features: list of dict
    :param label_col: The column which should contain the label.
    :type label_col: str
    :param relation: The name to give the relationship represented in this
                     featureset.
    :type relation: str
    :param arff_regression: Make the class variable numeric instead of nominal
                            and remove all non-numeric attributes.
    :type relation: bool
    '''
    # Convert IDs, classes, and features into format for DictWriter
    fields, instances = _examples_to_dictwriter_inputs(ids, classes, features,
                                                       label_col)
    if label_col in fields:
        fields.remove(label_col)

    # a list to keep track of any non-numeric fields
    non_numeric_fields = []

    # Write file
    if sys.version_info >= (3, 0):
        file_mode = 'w'
    else:
        file_mode = 'wb'
    with open(path, file_mode) as f:
        # Add relation to header
        print("@relation '{}'\n".format(relation), file=f)

        # Loop through fields writing the header info for the ARFF file
        sorted_fields = sorted(fields)
        for field in sorted_fields:
            # Check field type
            numeric = True
            for instance in instances:
                if field in instance:
                    numeric = not isinstance(instance[field], string_types)
                    break

            # ignore non-numeric fields if we are dealing with regression
            # but keep track of them for later use
            if arff_regression and not numeric:
                non_numeric_fields.append(field)
                continue

            print("@attribute '{}'".format(field.replace('\\', '\\\\')
                                                .replace("'", "\\'")),
                  end=" ", file=f)
            print("numeric" if numeric else "string", file=f)

        # Print class label header if necessary
        if arff_regression:
            print("@attribute {} numeric".format(label_col), file=f)
        else:
            if set(classes) != {None}:
                print("@attribute {} ".format(label_col) +
                      "{" + ','.join(map(str, sorted(set(classes)))) + "}",
                      file=f)
        sorted_fields.append(label_col)

        # throw out any non-numeric fields if we are writing for regression
        if arff_regression:
            sorted_fields = [fld for fld in sorted_fields if fld not in
                             non_numeric_fields]

        # Create CSV writer to handle missing values for lines in data section
        # and to ignore the instance values for non-numeric attributes
        writer = csv.DictWriter(f, sorted_fields, restval=0,
                                extrasaction='ignore', dialect='arff')

        # Output instances
        print("\n@data", file=f)
        writer.writerows(instances)


def _write_delimited_file(path, ids, classes, features, label_col, dialect):
    '''
    Writes a feature file in .csv or .tsv format with the given a list of IDs,
    classes, and features.

    :param path: A path to the feature file we would like to create.
    :type path: str
    :param ids: The IDs for each instance in the feature list/array. If None,
                IDs will be automatically generated with the prefix specified by
                `id_prefix` followed by the row number. If `id_prefix` is also
                None, no IDs will be written to the file.
    :type ids: list of str
    :param classes: The class labels for each instance in the feature
                    list/array. If None, no class labels will be added to output
                    file.
    :type classes: list of str
    :param features: The features for each instance.
    :type features: list of dict
    :param label_col: The column which should contain the label.
    :type label_col: str
    :param dialect: The CSV dialect to write the output file in. Should be
                    'excel-tab' for .tsv and 'excel' for .csv.
    :type dialect: str
    '''

    # Convert IDs, classes, and features into format for DictWriter
    fields, instances = _examples_to_dictwriter_inputs(ids, classes, features,
                                                       label_col)

    # Write file
    if sys.version_info >= (3, 0):
        file_mode = 'w'
    else:
        file_mode = 'wb'
    with open(path, file_mode) as f:
        # Create writer
        writer = DictWriter(f, fieldnames=sorted(fields), restval=0,
                            dialect=dialect)
        # Output instance
        writer.writeheader()
        writer.writerows(instances)


def _write_jsonlines_file(path, ids, classes, features):
    '''
    Writes a feature file in .jsonlines/.ndj format with the given a list of
    IDs, classes, and features.

    :param path: A path to the feature file we would like to create.
    :type path: str
    :param ids: The IDs for each instance in the feature list/array. If None,
                IDs will be automatically generated with the prefix specified by
                `id_prefix` followed by the row number. If `id_prefix` is also
                None, no IDs will be written to the file.
    :type ids: list of str
    :param classes: The class labels for each instance in the feature
                    list/array. If None, no class labels will be added to output
                    file.
    :type classes: list of str
    :param features: The features for each instance.
    :type features: list of dict
    '''
    file_mode = 'w' if sys.version_info >= (3, 0) else 'wb'
    with open(path, file_mode) as f:
        # Iterate through examples
        for ex_id, class_name, feature_dict in zip(ids, classes, features):
            # Don't try to add class column if this is label-less data
            example_dict = {}
            if class_name is not None:
                example_dict['y'] = class_name
            if ex_id is not None:
                example_dict['id'] = ex_id
            example_dict["x"] = feature_dict
            print(json.dumps(example_dict, sort_keys=True), file=f)


def _write_megam_file(path, ids, classes, features):
    '''
    Writes a feature file in .megam format with the given a list of IDs,
    classes, and features.

    :param path: A path to the feature file we would like to create.
    :type path: str
    :param ids: The IDs for each instance in the feature list/array. If None,
                IDs will be automatically generated with the prefix specified by
                `id_prefix` followed by the row number. If `id_prefix` is also
                None, no IDs will be written to the file.
    :type ids: list of str
    :param classes: The class labels for each instance in the feature
                    list/array. If None, no class labels will be added to output
                    file.
    :type classes: list of str
    :param features: The features for each instance.
    :type features: list of dict
    '''
    with open(path, 'w') as f:
        # Iterate through examples
        for ex_id, class_name, feature_dict in zip(ids, classes, features):
            # Don't try to add class column if this is label-less data
            if ex_id is not None:
                print('# {}'.format(ex_id), file=f)
            if class_name is not None:
                print(class_name, end='\t', file=f)
            print(' '.join(('{} {}'.format(field, value) for field, value in
                            sorted(feature_dict.items()) if
                            Decimal(value) != 0)),
                  file=f)


def _write_sub_feature_file(path, ids, classes, features, filter_features,
                            label_col='y', arff_regression=False,
                            arff_relation='skll_relation'):
    '''
    Writes a feature file in either ``.arff``, ``.csv``, ``.jsonlines``,
    ``.megam``, ``.ndj``, or ``.tsv`` formats with the given a list of IDs,
    classes, and features.

    :param path: A path to the feature file we would like to create. The suffix
                 to this filename must be ``.arff``, ``.csv``, ``.jsonlines``,
                 ``.megam``, ``.ndj``, or ``.tsv``
    :type path: str
    :param ids: The IDs for each instance in the feature list/array. If None,
                IDs will be automatically generated with the prefix specified by
                `id_prefix` followed by the row number. If `id_prefix` is also
                None, no IDs will be written to the file.
    :type ids: list of str
    :param classes: The class labels for each instance in the feature
                    list/array. If None, no class labels will be added to output
                    file.
    :type classes: list of str
    :param features: The features for each instance represented as either a list
                     of dictionaries.
    :type features: list of dict
    :param filter_features: Set of features to include in current feature file.
    :type filter_features: set of str
    :param label_col: Name of the column which contains the class labels for
                      CSV/TSV files. If no column with that name exists, or
                      `None` is specified, the data is considered to be
                      unlabelled.
    :type label_col: str
    :param arff_regression: A boolean value indicating whether the ARFF files
                            that are written should be written for arff_regression
                            rather than classification, i.e., the class variable
                            y is numerical rather than an enumeration of classes
                            and all non-numeric attributes are removed.
    :type label_col: bool
    :param arff_relation: Relation name for ARFF file.
    :type label_col: str
    :param subsets: A mapping from subset names to lists of feature names that
                    are included in those sets. If given, a feature file will
                    be written for every subset (with the name containing the
                    subset name as suffix to ``path``).
    :type subsets: dict (str to list of str)
    '''
    # Setup logger
    logger = logging.getLogger(__name__)

    # Get lowercase extension for file extension checking
    ext = os.path.splitext(path)[1].lower()

    # Filter feature dictionaries if asked
    if filter_features:
        # Print pre-filtered list of feature names if debugging
        if logger.getEffectiveLevel() == logging.DEBUG:
            feat_names = set()
            for feat_dict in features:
                feat_names.update(feat_dict.keys())
            feat_names = sorted(feat_names)
            logger.debug('Original features: %s', feat_names)

        # Apply filtering
        features = [{feat_name: feat_value for feat_name, feat_value in
                     iteritems(feat_dict) if (feat_name in filter_features or
                                              feat_name.split('=', 1)[0] in
                                              filter_features)} for feat_dict
                    in features]

        # Print list post-filtering
        if logger.getEffectiveLevel() == logging.DEBUG:
            feat_names = set()
            for feat_dict in features:
                feat_names.update(feat_dict.keys())
            feat_names = sorted(feat_names)
            logger.debug('Filtered features: %s', feat_names)

    # Create TSV file if asked
    if ext == ".tsv":
        _write_delimited_file(path, ids, classes, features, label_col,
                              'excel-tab')
    # Create CSV file if asked
    elif ext == ".csv":
        _write_delimited_file(path, ids, classes, features, label_col,
                              'excel')
    # Create .jsonlines file if asked
    elif ext == ".jsonlines" or ext == '.ndj':
        _write_jsonlines_file(path, ids, classes, features)
    # Create .megam file if asked
    elif ext == ".megam":
        _write_megam_file(path, ids, classes, features)
    # Create ARFF file if asked
    elif ext == ".arff":
        _write_arff_file(path, ids, classes, features, label_col,
                         arff_regression=arff_regression,
                         relation=arff_relation)
    # Invalid file suffix, raise error
    else:
        raise ValueError(('Output file must be in either .arff, .csv, ' +
                          '.jsonlines, .megam, .ndj, or .tsv format. You ' +
                          'specified: {}').format(path))


def write_feature_file(path, ids, classes, features, feat_vectorizer=None,
                       id_prefix='EXAMPLE_', label_col='y',
                       arff_regression=False, arff_relation='skll_relation',
                       subsets=None):
    '''
    Writes a feature file in either ``.arff``, ``.csv``, ``.jsonlines``,
    ``.megam``, ``.ndj``, or ``.tsv`` formats with the given a list of IDs,
    classes, and features.

    :param path: A path to the feature file we would like to create. The suffix
                 to this filename must be ``.arff``, ``.csv``, ``.jsonlines``,
                 ``.megam``, ``.ndj``, or ``.tsv``. If ``subsets`` is not
                 ``None``, this is assumed to be a string containing the path to
                 the directory to write the feature files with an additional
                 file extension specifying the file type. For example
                 ``/foo/.csv``.
    :type path: str
    :param ids: The IDs for each instance in the feature list/array. If None,
                IDs will be automatically generated with the prefix specified by
                `id_prefix` followed by the row number. If `id_prefix` is also
                None, no IDs will be written to the file.
    :type ids: list of str
    :param classes: The class labels for each instance in the feature
                    list/array. If None, no class labels will be added to output
                    file.
    :type classes: list of str
    :param features: The features for each instance represented as either a list
                     of dictionaries or an array-like (if `feat_vectorizer`
                     is also specified).
    :type features: list of dict or array-like
    :param feat_vectorizer: A `DictVectorizer` to map to/from feature columns
                            indices and names.
    :type feat_vectorizer: DictVectorizer
    :param id_prefix: If we need to automatically generate IDs, put this prefix
                      infront of the numerical IDs.
    :type id_prefix: str
    :param label_col: Name of the column which contains the class labels for
                      CSV/TSV files. If no column with that name exists, or
                      `None` is specified, the data is considered to be
                      unlabelled.
    :type label_col: str
    :param arff_regression: A boolean value indicating whether the ARFF files
                            that are written should be written for regression
                            rather than classification, i.e., the class variable
                            y is numerical rather than an enumeration of classes
                            and all non-numeric attributes are removed.
    :type arff_regression: bool
    :param arff_relation: Relation name for ARFF file.
    :type arff_relation: str
    :param subsets: A mapping from subset names to lists of feature names that
                    are included in those sets. If given, a feature file will
                    be written for every subset (with the name containing the
                    subset name as suffix to ``path``). Note, since
                    string-valued features are automatically converted into
                    boolean features with names of the form
                    ``FEATURE_NAME=STRING_VALUE``, when doing the filtering, the
                    portion before the ``=`` is all that's used for matching.
                    Therefore, you do not need to enumerate all of these boolean
                    feature names in your mapping.
    :type subsets: dict (str to list of str)
    '''
    # Setup logger
    logger = logging.getLogger(__name__)

    logger.debug('Feature vectorizer: %s', feat_vectorizer)
    logger.debug('Features: %s', features)

    # Check for valid features
    if feat_vectorizer is None and features and not isinstance(features[0],
                                                               dict):
        raise ValueError('If `feat_vectorizer` is unspecified, you must pass ' +
                         'a list of dicts for `features`.')

    # Convert features to list of dicts if given an array-like and a vectorizer
    if feat_vectorizer is not None:
        features = feat_vectorizer.inverse_transform(features)

    # Create ID generator if necessary
    if ids is None:
        if id_prefix is not None:
            ids = ('{}{}'.format(id_prefix, num) for num in
                   range(len(features)))
        else:
            ids = (None for _ in range(len(features)))

    # Create class generator if necessary
    if classes is None:
        classes = (None for _ in range(len(features)))

    # Get prefix and extension for checking file types and writing subset files
    root, ext = re.search(r'^(.*)(\.[^.]*)$', path).groups()

    # Write one feature file if we weren't given a dict of subsets
    if subsets is None:
        _write_sub_feature_file(path, ids, classes, features, [],
                                label_col=label_col,
                                arff_regression=arff_regression,
                                arff_relation=arff_relation)
    # Otherwise write one feature file per subset
    else:
        ids = list(ids)
        classes = list(classes)
        for subset_name, filter_features in iteritems(subsets):
            logger.debug('Subset (%s) features: %s', subset_name,
                         filter_features)
            sub_path = os.path.join(root, '{}{}'.format(subset_name, ext))
            _write_sub_feature_file(sub_path, ids, classes, features,
                                    set(filter_features), label_col=label_col,
                                    arff_regression=arff_regression,
                                    arff_relation=arff_relation)
